; Codeschloss
; ======================================================================================
; Als Eingnge fr die Codetasten sind zwei Ports KEY_BANKA und KEY_BANKB (Tastenbank) vorgesehen:
; - KEY_BANKA = Codetasten 1 bis 8 auf den Portpins 0 bis 7
; - KEY_BANKB = Programmierkontakt auf einem Portpin, restliche 7 Portpins fr weitere Codetasten, falls erforderlich
; Die Definition von KEY_BANKA, KEY_BANKB und der Pins stehen in den Konstanten weiter unten.
; Je nach Nutzung der Pins von KEY_BANKB als Ein- oder Ausgang ist die Aktivierung der Pullup-Widerstnde anzupassen

; Die Ausgnge befinden sich alle auf einer Bank (diese kann mit KEY_BANKB identisch sein):
; - Trffner
; - LED (rot) fr Anzeige der laufenden Codeeingabe (leuchtet, solange nur richtige Codetaste gedrckt wurden)
; - LED (gelb) zur Anzeige des aktiven Code-Programmiermodus; blinkt, wenn ffnungsdauer programmiert wird
; - LED (grn) zur Anzeige des gerade aktiven Trffners
; Die Pins fr Trffner und LEDs sind weiter unten als Konstanten definiert.

; (*) im Kommentar einer Zeile zeigt ggf. anzupassende Stellen an, falls andere Ports verwendet werden sollen

; Zwei Zielplattformen sind vorgesehen:
; - ATmega8 zur Entwicklung auf dem myAVR-Testboard (Clock-Fuse auf 1 MHz internen Takt)
; - ATtiny2313 zum Einsatz auf eigener Schaltung (Clock-Fuse auf internen 4 MHz-Takt)
; Steuerung erfolgt durch wechselweises Auskommentieren einer der folgenden Zeilen:

; Zielplattform definieren =============================================================
	 .SET TINY = 1				; Assemblieren fr ATtiny2313
;	.SET MEGA = 1				; Assemblieren fr ATmega8
; ======================================================================================

; Matrixtastatur erzwingen =============================================================
;	.SET FORCE_MATRIX_KEYPAD = 1		; nur Matrixtastatur untersttzen
; ======================================================================================

	.NOLIST

; Definitionen fr ATmega8 =============================================================

	.IFDEF MEGA				; Include-Datei fr ATmega8

	.INCLUDE "m8def.inc"

	.equ CLOCK = 1000000			; Takt des Controllers (ber Fuse-Bit eingestellter Wert)
	.equ CKDIV8_VALUE = 1			; CKDIV8-Fuse gibt es bei ATmega8 nicht, also 1
	.equ INT_PRESCALE = 64			; weiter unten eingestellter Interrupt-Teiler
	.equ PRESCALER = (1 << CS01) | (1 << CS00)

	.ENDIF

; Definitionen fr ATtiny2313 ==========================================================

	.IFDEF TINY				; Include-Datei fr ATtiny2313

	.INCLUDE "tn2313def.inc"

	.equ CLOCK = 4000000			; Takt des Controllers (ber Fuse-Bit eingestellter Wert)
	.equ CKDIV8_VALUE = 8			; 8 wenn CKDIV8-Fuse gesetzt, sonst 1
	.equ INT_PRESCALE = 64			; weiter unten eingestellter Interrupt-Teiler
	.equ PRESCALER = (1 << CS01) | (1 << CS00)

	.ENDIF

	.LIST
	.LISTMAC

; Folgende Angaben dienen zur Berechnung des Timings ===================================

	; Startwert fr Interruptzhler errechnen
	; Interrupt-Teilung auf etwa 0.2 Sekunden
	.equ INT_COUNTER_INIT = INT((0.2 * CLOCK) / (CKDIV8_VALUE * 256 * INT_PRESCALE))	

	.equ STARTUP_COUNTDOWN_INIT = 6 	; Dauer der Startphase des Codeschlosses in Anzahl Blinkzyklen
						; In der Startphase blinken alle LEDs als Lebenszeichen drei Mal

	; Definition der Timer-Startwerte
	.equ DOOR_TIMER_DEFAULT = 9		; Vorgabewert: Timeout fr Trffner (Wert im EEPROM)
	.equ SCAN_TIMEOUT_DEFAULT = 10		; Vorgabewert: Timeout der Codeeingabe (in Interruptzyklen, Wert im EEPROM)
	.equ INT_LED_FLASH_INIT = 2		; Verzgerung fr das Blinken der Programmier-LED


	; reservierte Tasten fr die Programmierung der Aktivierdauer
	.equ PROG_DOOR_TEST = 0			; Taste fr testweise Aktivierung (*)
	.equ PROG_DOOR_PLUS = 1			; Taste fr Aktivierdauer verlngern (*)
	.equ PROG_DOOR_MINUS = 2		; Taste fr Aktivierdauer verkrzen (*)

	; Mamimalwert fr Aktivierdauer des Trffners
	.equ MAX_DOOR_OPEN_TIME = 20


	.IF INT_COUNTER_INIT > 255
		.ERROR "CLOCK und CKDIV8 berprfen: Teiler in Interruptroutine ist nicht als 8-Bit-Zahl darstellbar."
	.ENDIF
	
; Registerdefinitionen =================================================================

	.def tmpr = R16				; Register fr temporre Daten
	.def counterr = R17 			; Register fr Zhlzwecke
	.def modusr = R18			; Modus, in dem sich das Codeschlosses gerade befindet
	.def doorcntr = R19			; Countdown fr Trffner
	.def scantimeoutr = R20			; Countdown fr die Zeit der Codeeingabe
	.def ledflashcntr = R21			; Countdown-Zhler fr LED-On-Off-Phasen, um Blinken zu erzeugen

	; Folgende Register werden im Interrupt verwendet und dort nicht gesichert
	; Register fr Tastenbank A

	.def keylog_A = R0			; Key logical 
	.def cnt0_A = R1			; vertical counter lo
	.def cnt1_A = R2			; vertical counter hi
	.def key_pressed_A  = R3		; gedrckte Tasten
	.def key_released_A = R4		; losgelassene Tasten
	.def keymask_A = R5			; Maske zum Ausblenden nicht verwendeter Pins

	; Register fr Tastenbank B
	.def keylog_B = R6			; Key logical 
	.def cnt0_B = R7			; vertical counter lo
	.def cnt1_B = R8			; vertical counter hi
	.def key_pressed_B  = R9		; gedrckte Tasten
	.def key_released_B = R10		; losgelassene Tasten
	.def keymask_B = R11			; Maske zum Ausblenden nicht verwendeter Pins

	.def keyphy_A = R12			; Key physical A
	.def keyphy_B = R13			; Key physical B
	.def keytemp = R14			; Key temporr

	.def keymatrix = R15			; eine Tastaturmatrix ist angeschlossen

; Definition der mglichen Zustnde des Codeschlosses ==================================

	.equ SCAN_MODE = 0			; Codeschloss nimmt Tastendrcke entgegen und prft diese
	.equ OPEN_MODE = 1			; Trffner ist gerade aktiv
	.equ PROG_CODE_MODE = 2			; Codeschloss ist im Code-Programmiermodus
	.equ PROG_DOOR_MODE = 3			; Programmiermodus fr die Aktivierdauer des Trffners
	.equ STARTUP_MODE = 4			; Startup-Modus fr LED-Animation (Lebenszeichen des Codeschlosses)


; Definition der Ports =================================================================

	.IFDEF MEGA				; Definitionen fr ATmega8
	
	; Tastenbank A (Eingang)
	.equ KEY_BANKA_PORT = PORTB	; PORT (*)
	.equ KEY_BANKA_PIN = PINB	; PIN  (*)
	.equ KEY_BANKA_DDR = DDRB	; DDR  (*)
	
	; Tastenbank B (Eingang)
	.equ KEY_BANKB_PORT = PORTC	; PORT (*)
	.equ KEY_BANKB_PIN = PINC	; PIN  (*)
	.equ KEY_BANKB_DDR = DDRC	; DDR  (*)

	; Programmierkontakt (Eingang)
	.equ PROG_PINNUM= 5		; PIN-Nummer des Programmierkontakts (*)
	
	; Muster, um nicht genutzte Eingnge zu neutralisieren (werden hinzuge-oder-t)
	; nur relevant, wenn mit Einzelkontakten (keine Matrix) gearbeitet wird
	.equ KEY_BANKA_MASK = 0b00000000	; (*)
	.equ KEY_BANKB_MASK = 0b00000000	; (*)


	; Definition der Ausgnge fr ATmega8

	; Trffner (Ausgang)
	.equ DOOR_PORT = PORTD		; (*)
	.equ DOOR_DDR = DDRD		; PORT fr Trffner (*)
	.equ DOOR_PINNUM = 5		; PIN-Nummer (*)

	; LED fr laufenden Codeeingabe (Ausgang)
	.equ LED_CODE_PORT = PORTD	; PORT (*)
	.equ LED_CODE_PINNUM = 2	; PIN-Nummer (*)

	; LED fr Programmierung (Ausgang)
	.equ LED_PROG_PORT = PORTD	; (*)
	.equ LED_PROG_PINNUM = 3	; (*)

	; LED fr aktiven Trffner (Ausgang)
	.equ LED_DOOR_PORT = PORTD	; (*)
	.equ LED_DOOR_PINNUM = 4	; (*)

	.ENDIF


	.IFDEF TINY					; Definitionen fr ATtiny2313

	; Tastenbank A (Eingang)
	.equ KEY_BANKA_PORT = PORTB	; PORT (*)
	.equ KEY_BANKA_PIN = PINB	; PIN  (*)
	.equ KEY_BANKA_DDR = DDRB	; DDR  (*)
	
	; Tastenbank B (Eingang)
	.equ KEY_BANKB_PORT = PORTD	; PORT (*)
	.equ KEY_BANKB_PIN = PIND	; PIN  (*)
	.equ KEY_BANKB_DDR = DDRD	; DDR  (*)

	; Programmierkontakt (Eingang)
	.equ PROG_PINNUM= 6		; PIN-Nummer des Programmierkontakts (*)

	; Muster, um nicht genutzte Eingnge zu neutralisieren (werden hinzuge-oder-t)
	; nur relevant, wenn mit Einzelkontakten (keine Matrix) gearbeitet wird
	.equ KEY_BANKA_MASK = 0b00000000	; (*)
	.equ KEY_BANKB_MASK = 0b10111111	; Pin 6 ist Programmierkontakt (*)


	; Definition der Ausgnge fr ATtiny2313

	; Trffner (Ausgang)
	.equ DOOR_PORT = PORTD		; (*)
	.equ DOOR_DDR = DDRD		; PORT fr Trffner (*)
	.equ DOOR_PINNUM = 0		; PIN-Nummer (*)

	; LED fr laufenden Codeeingabe (Ausgang)
	.equ LED_CODE_PORT = PORTD	; PORT (*)
	.equ LED_CODE_PINNUM = 1	; PIN-Nummer (*)

	; LED fr Programmierung (Ausgang)
	.equ LED_PROG_PORT = PORTD	; (*)
	.equ LED_PROG_PINNUM = 2	; (*)

	; LED fr aktiven Trffner (Ausgang)
	.equ LED_DOOR_PORT = PORTD	; (*)
	.equ LED_DOOR_PINNUM = 3	; (*)
	
	.ENDIF

; Programmanfang ======================================================================= 

	.CSEG					; Codesegment
	.ORG 0x0000				; Vektortabelle

	rjmp RESET				; Reset
	
	.org OVF0addr
    	rjmp TIMER0_OVERFLOW	    		; Timer0 Overflow Handler

	reti

; ======================================================================================
; Hauptprogramm ========================================================================
; ======================================================================================

RESET:

	wdr					; vorsichtshalber Watchdog resetten
	in tmpr, WDTCR				; Timed Sequence fr Watchdog-Konfiguration einhalten
	ori tmpr, (1 << WDCE) | (1 << WDE)
	out WDTCR, tmpr
	ldi tmpr, (1 << WDE) | (1 << WDP2) | (1 << WDP1) | (1 << WDP0) 
	out WDTCR, tmpr				; Watchdog Timer Control Register auf lngste Zeit stellen

	; Stack initialisieren
	ldi tmpr, LOW(RAMEND)   		; LOW-Byte der obersten RAM-Adresse
	out SPL, tmpr
	
	.IFDEF MEGA				; High-Byte-gibt es bei ATtiny2313 nicht
	ldi tmpr, HIGH(RAMEND)  		; HIGH-Byte der obersten RAM-Adresse
	out SPH, tmpr
	.ENDIF

INIT:
	; Initialisieren

	.IFDEF MEGA				; Definitionen fr ATmega8

	ldi tmpr, 0x00				; Ports fr
	out KEY_BANKA_DDR, tmpr			; KEY_BANKA und
	out KEY_BANKB_DDR, tmpr			; KEY_BANKB komplett als Eingang (*)

	ldi tmpr, 0xFF
	out KEY_BANKA_PORT, tmpr		; Pullup-Widerstnde an Eingangsports B und C einschalten (*)
	out KEY_BANKB_PORT, tmpr		; (*)

	ldi tmpr, 0xFF				; OUT_PORT, alle Pins auf Ausgang
	out DOOR_DDR, tmpr			; (*)

	.ENDIF

	.IFDEF TINY				; Definitionen fr ATtiny2313

	ldi tmpr, 0x00
	out KEY_BANKA_DDR, tmpr			; KEY_BANKA komplett als Eingang

	ldi tmpr, 0b10111111			; (*)
	out KEY_BANKB_DDR, tmpr			; auf KEY_BANKB nur PIN 6 als Eingang fr Programmierkontakt

	com tmpr				; Pullup-Widerstand an KEY_BANKB nur an Port 6 (Programmierkontakt) (*)
	out KEY_BANKB_PORT, tmpr

	ldi tmpr, 0xFF
	out KEY_BANKA_PORT, tmpr		; Pullup-Widerstnde an KEY_BANKA einschalten

	.ENDIF

	ldi tmpr, PRESCALER			; Timer 0 mit Teiler PRESCALER bestcken
	out TCCR0, tmpr

	ldi tmpr, (1 << TOIE0)			; Timer-Interrupt 0 einschalten
	out TIMSK, tmpr

	; Tastenstatus initialisieren fr Keybank A
	ldi tmpr, KEY_BANKA_MASK		; Keymask fr Tastenbank A laden
	mov keymask_A, tmpr
	in keylog_A, KEY_BANKA_PIN		; logische Tastenzustnde von den physischen bernehmen
	or keylog_A, keymask_A
	clr cnt0_A				; vertical counter low
	clr cnt1_A				; vertical counter high

	; Tastenstatus initialisieren fr Keybank B
	ldi tmpr, KEY_BANKB_MASK		; Keymask fr Tastenbank B laden
	mov keymask_B, tmpr
	in keylog_B, KEY_BANKB_PIN		; logische Tastenzustnde von den physischen bernehmen
	or keylog_B, keymask_B
	clr cnt0_B				; vertical counter low
	clr cnt1_B				; vertical counter high

	clr keymatrix				; Default ist "keine Matrixtastatur"
	
	.IFNDEF FORCE_MATRIX_KEYPAD		; wenn FORCE_MATRIX definiert, folgenden Block nicht assemblieren, damit Matrixtastatur erzwingen
	sbic KEY_BANKA_PIN, 7			; wenn Pin 7 geschlossen, dann ist Matrixtastatur angeschlossen
 	rjmp NO_KEYMATRIX_DETECTED		; ansonsten einzelne Tasten
	.ENDIF

	com keymatrix				; Matrixtastatur ist angeschlossen, in keymatrix merken (= 0xFF)

; Ports fr Matrixtastatur konfigurieren ===============================================

	ldi tmpr, 0b01110000			; Pins 0 bis 3 als Eingang, 4 bis 6 als Ausgang, 7 bleibt Eingang
	out KEY_BANKA_DDR, tmpr			; KEY_BANKA komplett als Eingang
	
	com tmpr
	out KEY_BANKA_PORT, tmpr		; Pullup-Widerstnde fr die Eingangspins einschalten


NO_KEYMATRIX_DETECTED:

	; Startphase initialisieren
	ldi counterr, INT_COUNTER_INIT		; Interrupt-Zhler mit errechnetem Startwert laden
	sts INT_COUNTER, counterr

	ldi tmpr, STARTUP_COUNTDOWN_INIT	; Dauer der Startphase des Codeschlosses in Anzahl Countdown-Zyklen
	sts STARTUP_COUNTDOWN, tmpr

	; Aktivierdauer des Trffners und Code-Eingabezeit aus EEPROM lesen
	ldi ZL, LOW(INIT_VALUES_EEPROM)
	ldi ZH, HIGH(INIT_VALUES_EEPROM)
	rcall READ_BYTE_FROM_EEPROM		; Aktivierdauer des Trffners lesen
	sts DOOR_OPEN_TIMER_INIT, tmpr		; und ins SRAM bernehmen
	adiw ZL, 1
	rcall READ_BYTE_FROM_EEPROM		; dasselbe fr die
	sts SCAN_TIMEOUT_INIT, tmpr		; Scan-Zeit fr Codeeingabe

	; Code aus EEPROM in SRAM laden
	ldi XL, LOW(CODE_TABLE_SRAM)
	ldi XH, HIGH(CODE_TABLE_SRAM)

	ldi ZL, LOW(CODE_TABLE_EEPROM)
	ldi ZH, HIGH(CODE_TABLE_EEPROM)

	ldi counterr, (CODE_TABLE_EEPROM_END-CODE_TABLE_EEPROM)
	sts CODE_TABLE_SIZE, counterr					; Gre der Codetabelle
	ldi counterr, (CODE_TABLE_EEPROM_END-CODE_TABLE_EEPROM)+2	; Gre der Codetabelle plus 0xFFFF-Endewort
	sts CODE_TABLE_INCL_ENDFLAG_SIZE, counterr

CODE_LOOP:					; Codetabelle einschlielich 0xFFFF-Endewort vom EEPROM ins SRAM kopieren
	rcall READ_BYTE_FROM_EEPROM
	st X+, tmpr
	adiw ZL, 1
	dec counterr
	brne CODE_LOOP

	ldi modusr, STARTUP_MODE		; mit Startup-Modus beginnen
	ldi ledflashcntr, INT_LED_FLASH_INIT	; Zhler fr das Blinken der LED laden

	sei					; Interrupts einschalten


; Hauptschleife ========================================================================
; ======================================================================================

MAIN_LOOP:

	wdr					; Watchdog zurcksetzen

	ldi ZL, LOW(JUMP_TABLE)			; Status-abhngig auf Handler-Routine verzweigen
	ldi ZH, HIGH(JUMP_TABLE)
	add ZL, modusr
	ldi tmpr, 0
	adc ZH, tmpr

	ijmp					; auf Routine gem Status verzweigen
 
JUMP_TABLE:					; Sprungtabelle auf Status-spezifische Routinen
	rjmp SCAN_CODE				; Reihenfolge synchron zu den x_MODE-Werten
	rjmp OPEN_DOOR
	rjmp PROG_CODE
	rjmp PROG_DOOR
	rjmp STARTUP

; Modusspezifische Routinen ============================================================
; ======================================================================================

SCAN_CODE:					; SCAN-Modus

	rcall WAIT_FOR_KEY_REL			; auf Loslassen einer Taste warten
	sbrs key_released_B, PROG_PINNUM	; auf Programmierkontakt prfen (auf Tastenbank B)
	rjmp SCAN_CODE_INPUT			; nein, zur Codeeingabe

	ldi modusr, PROG_CODE_MODE		; auf Code-Programmiermodus wechseln
	ldi YL, LOW(CODE_TABLE_SRAM)		; Adresse der Codetabelle im SRAM laden 
	ldi YH, HIGH(CODE_TABLE_SRAM)
	lds counterr, CODE_TABLE_SIZE		; Gre der Codetabelle

	rjmp MAIN_LOOP

SCAN_CODE_INPUT:				; Taste wurde losgelassen
	ld tmpr, Y+				; erwarteter Teil 1 der Taste aus SRAM
	eor key_released_A, tmpr		; EOR-verknpft mit losgelassenem Teil 1 muss Null ergeben (1)
	
	ld tmpr, Y+				; erwarteter Teil 2 der Taste aus SRAM

	eor key_released_B, tmpr		; EOR-verknpft mit losgelassenem Teil 2 muss Null ergeben (2)
	or key_released_B, key_released_A	; Teil (1) und Teil (2) mssen Oder-verknpft Null sein fr richtige Taste

	breq SCAN_KEY_OK			; Taste war richtig

	rcall PREPARE_SCAN_MODE			; Taste war falsch, zurck auf Anfang
	rjmp MAIN_LOOP

SCAN_KEY_OK:
	sbi LED_CODE_PORT, LED_CODE_PINNUM	; LED fr laufende Codeeingabe ein
	lds scantimeoutr, SCAN_TIMEOUT_INIT	; Timeout fr Codeeingabe neu laden
	ld tmpr, Y				; Endekennzeichen erreicht?
	ldd counterr, Y+1			; Hierzu mssen beide Bytes in Codetabelle
 	and tmpr, counterr			; gleich 0xFF sein, counterr kurzfristig zweckentfremden
	com tmpr				; 0xFF umdrehen zu 0x00
	brne MAIN_LOOP				; nein, weiter im SCAN-Modus bleiben

	; Code wurde richtig eingegeben, Trffner aktivieren
	ldi modusr, OPEN_MODE			; auf Trffner-aktiv-Modus wechseln
	rcall SET_DOOR_ON

	rjmp MAIN_LOOP

; ======================================================================================

OPEN_DOOR:					; Trffner aktiv
	cli					; Interrupts deaktivieren fr sauberen Status
	clr key_pressed_A
	clr key_pressed_B
	clr key_released_A
	clr key_released_B
	sei

DOOR_LOOP:
	wdr
	mov tmpr, key_pressed_A
	or tmpr, key_pressed_B
	brne OPEN_DOOR_CANCEL_KEY		; wenn eine Taste gedrckt wurde, Trffner abschalten
	tst doorcntr				; sonst: ist Zeit fr den Trffner abgelaufen?
	brne DOOR_LOOP				; ja, OPEN-Modus ist beendet
	rjmp OPEN_DOOR_MODE_END			; nein, zurck in Hauptschleife

OPEN_DOOR_CANCEL_KEY:
	rcall PREPARE_SCAN_MODE

WAIT_FOR_KEY_RELEASE:
	wdr
	mov tmpr, key_released_A
	or tmpr, key_released_B			; warten, bis Tasten, mit der Trffner abgeschaltet wurde,
	breq WAIT_FOR_KEY_RELEASE		; wieder losgelassen wurden - wrden sonst als neue Codeeingabe interpretiert

OPEN_DOOR_MODE_END:
	rcall PREPARE_SCAN_MODE			; wieder auf SCAN-Mode stellen
	
	rjmp MAIN_LOOP

; ======================================================================================

PROG_CODE:					; Code-Programmierung

	cbi LED_CODE_PORT, LED_CODE_PINNUM	; LED fr laufende Codeeingabe ausschalten
	sbi LED_PROG_PORT, LED_PROG_PINNUM	; LED fr Programmiermodus einschalten

	rcall WAIT_FOR_KEY_REL			; auf Loslassen einer Taste warten

	sbrs key_released_B, PROG_PINNUM	; auf Programmierkontakt prfen (auf Tastenbank B)
	rjmp PROG_CODE_INPUT			; nein, zur Codeprogrammierung

	rjmp PROG_TERMINATE			; ja, Programmierung beenden
								
PROG_CODE_INPUT:
	st Y+, key_released_A			; gedrckte Taste der Tastenbank A speichern 
	st Y+, key_released_B			; gedrckte Taste der Tastenbank B speichern
	subi counterr, 2 			; um zwei herunterzhlen, da jede Codetaste zwei Bytes belegt
	
	brne PROG_CONTINUE			; bei Null sind Speichermglichkeiten erschpft; nicht Null: Programmierung fortsetzen

PROG_TERMINATE:					; Programmierung abschlieen

	lds tmpr, CODE_TABLE_SIZE
	cp tmpr, counterr			; wurden berhaupt neue Tasten gelernt?
	brne PROG_SAVE				; ja, Code ins EEPROM sichern und Programmiermodus beenden

	ldi modusr, PROG_DOOR_MODE		; kein Code eingegeben, Wechsel in Programmierung der Aktivierdauer
	cbi LED_PROG_PORT, LED_PROG_PINNUM	; LED fr Programmiermodus aus
	ldi ledflashcntr, INT_LED_FLASH_INIT	; Zhler fr das Blinken der LED laden
	
	rjmp MAIN_LOOP				; des Trffners

PROG_SAVE:					; Sichern der gelernten Tasten im EEPROM
	ldi counterr, 0xFF			; Endekennzeichen ist 0xFF
	st Y+, counterr				; setzen fr unteres
	st Y, counterr				; und oberes Byte einer Taste

	ldi XL, LOW(CODE_TABLE_SRAM)		; XL/XH mit Adresse auf Codetabelle im SRAM laden
	ldi XH, HIGH(CODE_TABLE_SRAM)
	ldi ZL, LOW(CODE_TABLE_EEPROM)		; ZL/ZH mit Adresse auf Codetabelle im EEPROM laden
	ldi ZH, HIGH(CODE_TABLE_EEPROM)
	lds counterr, CODE_TABLE_INCL_ENDFLAG_SIZE	; Gre der Ccdetabelle mit Endekennzeichen

PROG_LOOP:
	ld tmpr, X+				; Byte aus SRAM-Codetabelle lesen
	rcall WRITE_BYTE_TO_EEPROM		; im EEPROM ablegen
	adiw ZL, 1				; Zeiger auf Position im EEPROM weiterstellen
	dec counterr
	brne PROG_LOOP

PROG_QUIT:
	rcall PREPARE_SCAN_MODE			; wieder auf SCAN-Mode stellen

PROG_CONTINUE:
	rjmp MAIN_LOOP

; ======================================================================================
 
PROG_DOOR:					; Programmierung der ffner-Aktivierdauer

	rcall WAIT_FOR_KEY_REL			; auf Loslassen einer Taste warten

	lds tmpr, DOOR_OPEN_TIMER_INIT		; aktuelle Aktivierdauer

	sbrs key_released_B, PROG_PINNUM	; auf Programmierkontakt prfen (auf Tastenbank B)
	rjmp PROG_DOOR_INPUT			; nein, zur Programmierung der Aktivierdauer

	rjmp PROG_DOOR_TERMINATE		; ja, Programmierung beenden

PROG_DOOR_INPUT:				; Programmierung der Aktivierdauer der ffners

	bst key_released_A, PROG_DOOR_TEST	; Taste fr Testauslsung bettigt?
	brtc PROG_DOOR_NOT_TEST			; nein

	rcall SET_DOOR_ON			; Trffner wird testhalber bettigt
	rjmp MAIN_LOOP

PROG_DOOR_NOT_TEST:
	bst key_released_A, PROG_DOOR_PLUS
	brtc PROG_DOOR_NO_PLUS_KEY
	inc tmpr				; Plus-Taste, Aktivierdauer erhhen 
	cpi tmpr, MAX_DOOR_OPEN_TIME		; Maximalwert fr Aktivierdauer ist 8
	brlt PROG_DOOR_QUIT			; Wert begrenzen
	ldi tmpr, MAX_DOOR_OPEN_TIME
	rjmp PROG_DOOR_QUIT

PROG_DOOR_NO_PLUS_KEY:
	bst key_released_A, PROG_DOOR_MINUS
	brts PROG_DOOR_MINUS_KEY		; Minus-Taste gedrckt

	ldi ZL, LOW(DOOR_OPEN_TIMER_INIT_EEPROM); andernfalls Programmierung der Aktivierdauer verwerfen
	ldi ZH, HIGH(DOOR_OPEN_TIMER_INIT_EEPROM)
	rcall READ_BYTE_FROM_EEPROM		; Aktivierdauer des Trffners wieder aus EEPROM lesen 
	sts DOOR_OPEN_TIMER_INIT, tmpr		; und ins SRAM bernehmen
		 
	rjmp PROG_DOOR_FINISHED			; hierfr muss Aktivierdauer gelesen werden

PROG_DOOR_MINUS_KEY:
	dec tmpr				; Minus-Taste wurde bettigt
	brne PROG_DOOR_QUIT			; wenn noch nicht Null erreicht, ok
	ldi tmpr, 1				; sonst eins als Mindestwert laden

PROG_DOOR_QUIT:
	sts DOOR_OPEN_TIMER_INIT, tmpr		; vernderte Aktivierdauer ins SRAM

	rjmp MAIN_LOOP

PROG_DOOR_TERMINATE:
	ldi ZL, LOW(DOOR_OPEN_TIMER_INIT_EEPROM)
	ldi ZH, HIGH(DOOR_OPEN_TIMER_INIT_EEPROM)
	rcall WRITE_BYTE_TO_EEPROM		; neuen Timerwert ins EEPROM schreiben

PROG_DOOR_FINISHED:
	rcall PREPARE_SCAN_MODE			; wieder auf SCAN-Mode stellen
	rjmp MAIN_LOOP

; ======================================================================================

STARTUP:					; Startmodus

	lds tmpr, STARTUP_COUNTDOWN
	tst tmpr				; Startmodus beendet?
	brne WAIT_FOR_STARTUP_END		; nein, zurck in Hauptschleife
	
	rcall PREPARE_SCAN_MODE

WAIT_FOR_STARTUP_END:

	rjmp MAIN_LOOP


; Interrupt-Routine ====================================================================
; ======================================================================================

TIMER0_OVERFLOW:				; Timer-Interrupt 0

	push tmpr				; Register sichern:
	push counterr				; tempr und counterr
	in tmpr, SREG				; Statusregister sichern
	push tmpr
	
TIMER_JQ:

	; hier beginnt die Timer-Jobqueue (TIMER-JQ)

; Tasten einlesen und entprellen =======================================================
	; keyphy: physikalischer Status der Tasten (evtl. prellend)
	; keylog: logischer Status der Tasten (entprellt)
	; keyphy und keylog werden zunchst ge-EOR-d, um von physikalisch zu logisch abweichende Tasten zu ermitteln
	; cnt0 und cnt1 bilden einen Vertikalzhler, der mit den abweichenden Tasten AND-verknpft wird
	; Nicht genderte Tasten setzen den Zhler somit wieder auf Null. Wenn der Zhler den Wert 2 erreicht
	; hat (d.h. Bit in cnt1 = 1), wird der physikalische Bitstatus in den logischen bernommen.
	; Somit muss der physikalische Status zwei mal in Folge vom logischen abweichen, damit er
	; als stabiler Status in den logischen Status bernommen wird.

	tst keymatrix				; ist eine Matrixtastatur angeschlossen?
	breq NO_KEYMATRIX			; nein, einzeln angeschlossene Tasten abfragen

; Matrixtastatur abfragen ==============================================================
	ldi tmpr, 0b11101111			; erste Spalte der Matrixtastatur auf LO ziehen
	out KEY_BANKA_PORT, tmpr
	in tmpr, KEY_BANKA_PIN			; und einlesen
	andi tmpr, 0x0F				; Ergebnis steht in der unteren 4 Bits
	mov keyphy_A, tmpr			; als physikalischen Tastenstatus A bernehmen

	ldi tmpr, 0b11011111			; zweite Spalte der Matrixtastatur auf LO ziehen
	out KEY_BANKA_PORT, tmpr
	in tmpr, KEY_BANKA_PIN			; und einlesen
	andi tmpr, 0x0F				; Ergebnis steht in der unteren 4 Bits
	swap tmpr					; obere und untere 4 Bits tauschen
	or keyphy_A, tmpr			; als physikalischen Tastenstatus A bernehmen
	
	ldi tmpr, 0b10111111			; dritte Spalte der Matrixtastatur auf LO ziehen
	out KEY_BANKA_PORT, tmpr
	in tmpr, KEY_BANKA_PIN			; und einlesen
	andi tmpr, 0x0F				; Ergebnis steht in der unteren 4 Bits
	mov keyphy_B, tmpr			; als physikalischen Tastenstatus B bernehmen

	in tmpr, KEY_BANKB_PIN			; Programmierkontakt noch einlesen
	andi tmpr, (1 << PROG_PINNUM)
	eor keyphy_B, tmpr			; und hinzuverknpfen

	rjmp PROCESS_KEYS

NO_KEYMATRIX:
	in keyphy_A, KEY_BANKA_PIN		; physikalischen Tastenport (Bank A) einlesen
	or keyphy_A, keymask_A			; nicht genutzte Tasten als offen betrachten (auf HI setzen)

	in keyphy_B, KEY_BANKB_PIN		; physikalischen Tastenport (Bank B) einlesen
	or keyphy_B, keymask_B			; nicht genutzte Tasten als offen betrachten (auf HI setzen)

PROCESS_KEYS:
	; Tasten der Bank A auswerten
	mov keytemp, keyphy_A			; wird noch gebraucht, s.u.

	eor keyphy_A, keylog_A			; genderte Tasten ermitteln

	eor cnt1_A, cnt0_A			; vertical oounter high increment
	com cnt0_A				; vertical counter low  increment
	
	and cnt0_A, keyphy_A			; vertical counter reset, wenn physikalischer und logischer
	and cnt1_A, keyphy_A			; Tastenstatus gleich sind

	eor keylog_A, cnt1_A			; Tastenstatus in logischen Tastenstatus bernehmen
	
	mov keyphy_A, keytemp			; keytemp wird nochmals gebraucht
	and keytemp, cnt1_A
	or key_released_A, keytemp		; losgelassene Tasten der Bank A

	com keyphy_A				; Tastenstatus umdrehen
	and keyphy_A, cnt1_A
	or key_pressed_A, keyphy_A		; gedrckte Tasten der Bank A


	; Tasten der Bank B auswerten
	mov keytemp, keyphy_B			; wird noch gebraucht, s.u.

	eor keyphy_B, keylog_B			; genderte Tasten ermitteln

	eor cnt1_B, cnt0_B			; vertical oounter high increment
	com cnt0_B				; vertical counter low  increment
	
	and cnt0_B, keyphy_B			; vertical counter reset, wenn physikalischer und logischer
	and cnt1_B, keyphy_B			; Tastenstatus gleich sind

	eor keylog_B, cnt1_B			; Tastenstatus in logischen Tastenstatus bernehmen
	
	mov keyphy_B, keytemp			; keytemp wird nochmals gebraucht
	and keytemp, cnt1_B
	or key_released_B, keytemp		; losgelassene Tasten der Bank B

	com keyphy_B				; Tastenstatus umdrehen
	and keyphy_B, cnt1_B
	or key_pressed_B, keyphy_B		; gedrckte Tasten der Bank B


; Timerschleife durchlaufen ============================================================
; ======================================================================================

	lds counterr, INT_COUNTER   		; Interrupt-Zhler aus SRAM laden
	dec counterr				; Schleife, um INT_COUNTER_INIT Durchlufe zu verzgern
	sts INT_COUNTER, counterr		; Interrupt-Zhler zurckschreiben 

	brne TIMER_INT_QUIT			; noch nicht Null, Ende

	ldi counterr, INT_COUNTER_INIT		; Interrupt-Zhler wieder mit Startwert laden
	sts INT_COUNTER, counterr

; Steuerung von lnger laufenden Vorgngen =============================================
; folgender Programmcode wird ca. alle 0,2 Sekunden ausgefhrt
; ======================================================================================

	lds counterr, STARTUP_COUNTDOWN		; Test auf Startup-Phase
	tst counterr				; Startup-Countdown grer Null?
	breq STARTUP_DONE			; ja, beendet

	; alle drei LEDs in Startup-Phase langsam blinken lassen	
	dec ledflashcntr			; Zhler fr LED-Blinken herunterzhlen
	brne TIMER_INT_QUIT

	ldi ledflashcntr, INT_LED_FLASH_INIT	

	dec counterr
	sts STARTUP_COUNTDOWN, counterr

	ldi counterr, (1 << LED_CODE_PINNUM) | (1 << LED_PROG_PINNUM) | (1 << LED_DOOR_PINNUM)
	in tmpr, LED_CODE_PORT
	eor tmpr, counterr
	out LED_CODE_PORT, tmpr

	rjmp TIMER_INT_QUIT			; Startup-Phase luft noch, Interruptroutine beenden

STARTUP_DONE:
	dec scantimeoutr			; Timeout fr Codeeingabe herunterzhlen
	brne JQ_PROG_LED_CONTROL		; noch nicht bei Null, Timeout fr Codeeingabe nicht weiter betrachten
	cpi modusr, SCAN_MODE			; gerade im Codeeingabemodus?
	brne JQ_PROG_LED_CONTROL		; nein, dann Timerablauf irrelevant
	com key_released_A			; sonst fehlerhafte Tasten unterschieben fr Tastenbank A (Loslassen der Tasten)
						; nicht bei Tastenbank B, da dort der Programmierkontakt!

JQ_PROG_LED_CONTROL:				; Programmier-LED blinkt durch Ein- und Ausschalten
	dec ledflashcntr 
	brne JQ_DOOR_CONTROL
	ldi ledflashcntr, INT_LED_FLASH_INIT
	cpi modusr, PROG_DOOR_MODE		; nur wenn im Programmiermodus fr die Aktivierdauer des Trffners
	brne JQ_DOOR_CONTROL

	ldi tmpr, (1 << LED_PROG_PINNUM)
	in counterr, LED_PROG_PORT
	eor counterr, tmpr			; LED umschalten
	out LED_PROG_PORT, counterr

JQ_DOOR_CONTROL:
	; Trffner steuern in Abhngigkeit von DOOR_COUNT
	tst doorcntr				; Countdown fr Trffner lesen
	breq TIMER_INT_QUIT			; wenn Null, Trffner nicht anfassen
	dec doorcntr				; Countdown des ffners dekrementieren
	brne TIMER_INT_QUIT			; wenn Null erreicht, ffner deaktivieren

DOOR_TIMEOUT:
	rcall SET_DOOR_OFF			; Trffner aus
	 

TIMER_INT_QUIT:
	pop tmpr				; Register wiederherstellen
	out SREG, tmpr				; Statusregister wiederherstellen
	pop counterr
	pop tmpr

	reti


; Unterprogrammme ======================================================================

SET_DOOR_ON:					; Trffner ein
	lds doorcntr, DOOR_OPEN_TIMER_INIT	; Trffner-Countdown initialisieren
	sbi DOOR_PORT, DOOR_PINNUM		; Trffner aktivieren
	sbi LED_DOOR_PORT, LED_DOOR_PINNUM	; LED fr Trffner ein
	ret

SET_DOOR_OFF:					; Trffner aus
	clr doorcntr				; Trffner-Countdown auf Null setzen
	cbi DOOR_PORT, DOOR_PINNUM		; Trffner deaktivieren
	cbi LED_DOOR_PORT, LED_DOOR_PINNUM	; LED fr Trffner aus
	ret

; ======================================================================================

PREPARE_SCAN_MODE:				; SCAN-Modus initialisieren
	rcall SET_DOOR_OFF			; Trffner ausschalten
	lds scantimeoutr, SCAN_TIMEOUT_INIT	; Timeout fr Codeeingabe initialisieren
	ldi modusr, SCAN_MODE			; Trschloss in Scan-Modus setzen (Default)
	cbi LED_CODE_PORT, LED_CODE_PINNUM	; LED fr laufende Codeeingabe ausschalten
	cbi LED_PROG_PORT, LED_PROG_PINNUM	; LED fr Programmierung ausschalten
	ldi YL, LOW(CODE_TABLE_SRAM)
	ldi YH, HIGH(CODE_TABLE_SRAM)
	ret

; ======================================================================================

WAIT_FOR_KEY_REL:				; auf Loslassen einer beliebigen Taste warten
	cli					; Interrupts abschalten fr sauberen Status
	clr key_released_A			; gelste Tasten lschen
	clr key_released_B
	sei					; Interrupts wieder zulassen
WAIT_FOR_KEY_REL_LOOP:
	wdr
	mov tmpr, key_released_A		; mindestens eine Taste losgelassen?
	or tmpr, key_released_B
	breq WAIT_FOR_KEY_REL_LOOP		; nein, warten
	ret

; ======================================================================================

READ_BYTE_FROM_EEPROM:				; Werte an Adresse Z (ZH/ZL) aus EEPROM lesen
	wdr
	sbic EECR, EEWE				
	rjmp READ_BYTE_FROM_EEPROM		; warten auf Ende des vorherigen Schreibzugriffs

	cli					; zeitkritische Aktion, Interrupts deaktivieren
	
	.IFDEF MEGA
	out EEARH, ZH
	out EEARL, ZL
	sbi EECR, EERE				; Lesen aktivieren
	in tmpr, EEDR				; Byte aus EEPROM lesen
	.ENDIF

	.IFDEF TINY
	out EEAR, ZL
	sbi EECR, EERE				; Lesen aktivieren
	in tmpr, EEDR				; Byte aus EEPROM lesen
	.ENDIF

	sei					; Interrupts wieder aktivieren
	ret

; ======================================================================================

WRITE_BYTE_TO_EEPROM:
	wdr
	sbic EECR, EEWE
	rjmp WRITE_BYTE_TO_EEPROM		; warten auf Ende des vorherigen Schreibzugriffs

	cli					; zeitkritische Aktion, Interrupts deaktivieren

	.IFDEF MEGA
	out EEARH, ZH
	out EEARL, ZL
	out EEDR, tmpr
	.ENDIF

	.IFDEF TINY
	out EEAR, ZL
	out EEDR, tmpr
	.ENDIF
	
	sbi EECR, EEMWE				; Schreiben vorbereiten
	sbi EECR, EEWE				; schreiben

	sei					; Interrupts wieder aktivieren
	ret

; SRAM =================================================================================

	.DSEG					; SRAM

INT_COUNTER:
	.BYTE 1					; Interrupt-Zhler, dient der Heruntertaktung auf ca. 0.2 Sekunden

STARTUP_COUNTDOWN:
	.BYTE 1					; Countdown der Startup-Phase (in Interrupts)

; die folgenden Werte werden beim Programmstart aus EEPROM von hier bernommen

DOOR_OPEN_TIMER_INIT:
	.BYTE 1					; Aktivierdauer des Trffners (in Interrupts)

SCAN_TIMEOUT_INIT:
	.BYTE 1					; Timeout der Codeeingabe (in Interrups)

; ======================================================================================

CODE_TABLE_SIZE:
	.BYTE 1					; Gre der Codetabelle in Bytes

CODE_TABLE_INCL_ENDFLAG_SIZE:
	.BYTE 1					; Gre der Codetabelle mit 0xFFFF-Endewort in Bytes

CODE_TABLE_SRAM:
;  1 Wort pro Taste zur Speicherung des Port-Bitmusters: 0b000000xx:0bxxxxxxxx
; 12 Tasten plus 0xFFFF-Word fr Ende ergeben 26 Bytes ((*) ggf. anpassen, je nach Anzahl speicherbarer Codetasten)
	.BYTE 12 * 2 + 2

; EEPROM ===============================================================================

	.ESEG					; EEPROM-Segment
	.ORG 0x0000

CODE_TABLE_EEPROM:
; Standard-Tastencodes
;  1 Wort pro Taste zur Speicherung des Port-Bitmusters: 0b000000xx:0bxxxxxxxx
; 12 Tasten mit Reserve, Nullword = Ende
	.DW 0x0001				;   1. Taste
	.DW 0x0002				;   2. Taste
	.DW 0xFFFF				;   3. Taste
	.DW 0xFFFF				;   4. Taste
	.DW 0xFFFF				;   5. Taste	
	.DW 0xFFFF				;   6. Taste
	.DW 0xFFFF				;   7. Taste
	.DW 0xFFFF				;   8. Taste
	.DW 0xFFFF				;   9. Taste
	.DW 0xFFFF				;  10. Taste
	.DW 0xFFFF				;  11. Taste
	.DW 0xFFFF				;  12. Taste

CODE_TABLE_EEPROM_END:
	.DW 0xFFFF				;  Endekennzeichen

; ======================================================================================

; Folgende Werte werden nacheinander eingelesen und mssen daher direkt
; in angegebener Reihenfolge hintereinander stehen
; Angaben zu Dauern werden in 0,2-Sekunden-Einheiten gezhlt

INIT_VALUES_EEPROM:

DOOR_OPEN_TIMER_INIT_EEPROM:	
	.DB DOOR_TIMER_DEFAULT			; Aktivierdauer des ffners (in Interruptzyklen)

SCAN_TIMEOUT_INIT_EEPROM:
	.DB SCAN_TIMEOUT_DEFAULT		; Timeout der Codeeingabe (in Interruptzyklen)

; ======================================================================================
